# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

class Builder(object):
  """Abstract base class for builder implementations that can execute a parsed BUILD target."""

  def __init__(self, ferror, root_dir):
    self.ferror = ferror
    self.root_dir = root_dir

  def build(self, target, args):
    """Subclasses must implement a BUILD target executor.  The value returned should be an int,
    0 indicating success and any other value indicating failure.

    target: the parsed target to build
    args: additional arguments to the builder backend"""

    pass
