
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: PythonLibrary(src/python/twitter/pants/BUILD:pants-packaged)

from setuptools import setup

setup(**
{   'classifiers': [   'Intended Audience :: Developers',
                       'License :: OSI Approved :: Apache Software License',
                       'Operating System :: OS Independent',
                       'Programming Language :: Python'],
    'description': "Twitter's pants build tool.",
    'entry_points': {   'console_scripts': [   'pants = twitter.pants.bin.pants_exe:main']},
    'install_requires': [   'twitter.common.quantity==0.1.0',
                            'twitter.common.python==0.3.1',
                            'twitter.common.util==0.1.0',
                            'twitter.common.config==0.1.0',
                            'pytest==2.3.5',
                            'Mako==0.4.0',
                            'twitter.common.decorators==0.1.0',
                            'Markdown==2.1.1',
                            'twitter.common.process==0.1.0',
                            'psutil==1.1.2',
                            'twitter.common.confluence==0.1.0',
                            'requests==2.0.0',
                            'elementtree==1.2.7-20070827-preview',
                            'Pygments==1.4',
                            'ansicolors==1.0.2',
                            'pytest-cov==1.6'],
    'license': 'Apache License, Version 2.0',
    'name': 'twitter.pants',
    'namespace_packages': ['twitter', 'twitter.common'],
    'package_data': {   'twitter.pants': [   'docs/targets_graph.dot',
                                             'docs/invoking.rst',
                                             'docs/howto_contribute.rst',
                                             'docs/pagerank.py',
                                             'docs/howto_develop.rst',
                                             'docs/3rdparty_jvm.rst',
                                             'docs/first_tutorial.rst',
                                             'GROUPS',
                                             'docs/first_concepts.rst',
                                             'docs/docs.rst',
                                             'docs/publish_via_git.sh',
                                             'docs/gen.py',
                                             'docs/tasks_graph_tsort_specs.dot',
                                             'docs/Makefile',
                                             'docs/conf.py',
                                             'docs/pants-logo.png',
                                             'docs/taskdev/falcon.gif',
                                             'docs/README.rst',
                                             'docs/taskdev/taskdev.asc',
                                             'docs/target_addresses.rst',
                                             'docs/pants-logo.ico',
                                             'docs/dev_tasks.rst',
                                             'docs/3rdparty_py.rst',
                                             'docs/targets_graph_gen.dot',
                                             'docs/migration.rst',
                                             'docs/publish.rst',
                                             'docs/build_files.rst',
                                             'docs/setup_repo.rst',
                                             'BUILD',
                                             'docs/3rdparty.rst',
                                             'README.md',
                                             'docs/_templates/layout.html',
                                             'OWNERS',
                                             'docs/ThriftDeps.rst',
                                             'docs/tasks_graph.dot',
                                             'docs/internals.rst',
                                             'docs/pants_essentials_techtalk.asc',
                                             'docs/from_maven.rst',
                                             'docs/credits.rst',
                                             'docs/python-readme.rst',
                                             'docs/install.rst',
                                             'docs/tasks_graph_tsort.dot',
                                             'docs/JVMProjects.rst',
                                             'docs/index.rst'],
                        'twitter.pants.java': ['bin/ng.py'],
                        'twitter.pants.java.distribution': [   'Version.java',
                                                               'README.md',
                                                               'Version.class'],
                        'twitter.pants.python': [   'BUILD.twitter',
                                                    'BUILD',
                                                    'README.md'],
                        'twitter.pants.tasks': [   'templates/eclipse/pydevproject-3.7.mk',
                                                   'templates/builddictionary/goals_reference.mk',
                                                   'templates/jar_publish/pom.mk',
                                                   'templates/idea/module-12.mk',
                                                   'templates/eclipse/debug-launcher-3.7.mk',
                                                   'files/eclipse/org.eclipse.core.resources.prefs',
                                                   'files/eclipse/org.eclipse.jdt.ui.prefs',
                                                   'templates/ivy_resolve/ivy.mk',
                                                   'templates/eclipse/project-3.7.mk',
                                                   'templates/jar_publish/ivysettings.mk',
                                                   'templates/eclipse/org.eclipse.jdt.core.prefs-3.7.mk',
                                                   'templates/eclipse/classpath-3.7.mk',
                                                   'templates/idea/project-12.mk',
                                                   'templates/builddictionary/page.mk',
                                                   'templates/eclipse/factorypath-3.7.mk'],
                        'twitter.pants.tasks.javadoc.assets': [   'javadoc.html',
                                                                  'favicon.ico',
                                                                  'index.html',
                                                                  'javadoc-single-package.html'],
                        'twitter.pants.version': ['OWNERS']},
    'package_dir': {   '': 'src'},
    'packages': [   'twitter',
                    'twitter.common',
                    'twitter.common.python',
                    'twitter.common.python.bin',
                    'twitter.pants',
                    'twitter.pants.base',
                    'twitter.pants.bin',
                    'twitter.pants.buildtimestats',
                    'twitter.pants.commands',
                    'twitter.pants.engine',
                    'twitter.pants.fs',
                    'twitter.pants.goal',
                    'twitter.pants.ivy',
                    'twitter.pants.java',
                    'twitter.pants.java.distribution',
                    'twitter.pants.java.jar',
                    'twitter.pants.net',
                    'twitter.pants.net.http',
                    'twitter.pants.process',
                    'twitter.pants.python',
                    'twitter.pants.scm',
                    'twitter.pants.targets',
                    'twitter.pants.tasks',
                    'twitter.pants.tasks.javadoc',
                    'twitter.pants.tasks.javadoc.assets',
                    'twitter.pants.tasks.python',
                    'twitter.pants.version',
                    'twitter.thrift',
                    'twitter.thrift.text'],
    'url': 'https://github.com/twitter/commons',
    'version': '0.0.14-dev0',
    'zip_safe': True}
)
